/* 
 * A conditional principal component analysis + k-winners-take-all simulator
 *
 * ProgramName: cpca_kwta.c 
 * Author: Shinichi ASAKAWA <asakawa@twcu.ac.jp>
 * $Id: cpca_kwta.c,v 1.4 2001/11/26 07:39:37 asakawa Exp asakawa $
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include "general.h"
#include "rand_gen.h"
#include "fileio_util.h"
#include "matrix_util.h"

/* ------------------------------------------------------------------------
   initial set of global variables
   ------------------------------------------------------------------------ */
int FROM=1;
int TO=10000;
int seed=0;
int EPOCH=200;

/* for Hebbian learning */
double error_limit = 0.00001;
double learning_ratio = 0.01;
double rand_range = 0.1;

/* for k-winners-take-all circuit */
double leakage = 0.01;
#define LEAKAGE 0.01
double k_bias = 1.0;
double n2k_limit = 0.001;
double n2k_threshold = 1.0;
int N_of_winners = 1;

int NPATS = 0;               /* Number of patterns to be learned */
char *input_file_name = NULL;
char *init_file_name = NULL;
matrix example;
matinfo Input_Info;
bool verbose=False;
bool very_verbose=False;

/* NOTE: 0-th neuron will be handled as a threshold value
   The actual number of neurons in each layer will be automatically
   settled by program in accordance with both the command line arguments
   and input file. */
int N_of_input_layer = 0;
int N_of_hidden_layer = 0;

struct NEURON {
    double output;              /* Output values of this neuron */
    int Ninp;                   /* Number of input neurons */
    double input;
    double *inp_wgt;            /* array of input weights */
    struct NEURON **inp_neuron; /* the pointer to the input neurons */
    double (*output_func)();    /* the pointer to the output function */
    double old_output;
    double state;
};
typedef struct NEURON neuron;

neuron *input, *hidden;

/* linear output function */
double linear_f(double state);
/* double Dlinear_f(double state);*/
double tanh_f(double state);

/* initialize a neuron which has N inputs */
void initialize_neuron(neuron *a, int N, double (*output_func)(), double range);

void calc_output(neuron *a);

void usage(char *prgname)
{
    fprintf(stderr, "### Program name: %s - A conditional PCA simulator\n", prgname);
    fprintf(stderr, "### Creator: Shinichi ASAKAWA <asakawa@twcu.ac.jp>\n");
    fprintf(stderr, "Usage: %s -hidden <int> -input <filename> [-k <int>] [-leakage <double>] [-k_bias <double>] [-n2k_limit <double>] [-init <initial_weight_file_name>] [-error <double>] [-learn <double>] [-epoch <int>] [-to <int>] [-seed <int>] [-rand_range <double>] [-verbose] [-very_verbose]\n", prgname);
    fprintf(stderr, "Arguments: \n");
    fprintf(stderr, "     hidden, Number of hidden layer\n");
    fprintf(stderr, "     input, input file name to be learned\n");
    fprintf(stderr, "default settings of options: \n");
    fprintf(stderr, "      learn = %g: learning ratio\n", learning_ratio);
    fprintf(stderr, "      error = %g: error limit\n", error_limit);
    fprintf(stderr, "      epoch = %d: epoch \n", EPOCH);
    fprintf(stderr, "      k = %d number of winners\n", N_of_winners);
    fprintf(stderr, "      leakage = %g\n", leakage);
    fprintf(stderr, "      k_bias = %g\n", k_bias);
    fprintf(stderr, "      n2k_limit = %g\n", n2k_limit);
    fprintf(stderr, "      n2k_threshold = %g\n", n2k_threshold);
    fprintf(stderr, "      to = %d: max iterations\n", TO);
    fprintf(stderr, "      rand_range: = %g: range for initial values\n", rand_range);
    fprintf(stderr, "other options: \n");
    fprintf(stderr, "      seed: seed of random number generator\n");
    fprintf(stderr, "      verbose: redundant outputs\n");
    fprintf(stderr, "      very_verbose: too much redundant outputs\n");
}

double linear_f(double state)
{
    return state;
}

double tanh_f(double state)
{
    return tanh(state);
}

/*
double Dlinear_f(double state)
{
    return 1.0;
}
*/
/* Initilazing a neuron, which is described as a struct NEURON */
/* in neuron_util.h, and assigning an initial value as default */
/* to each variable  */
void initialize_neuron(neuron *a, int N, double output_func(), double range )
{
    int i;

    a->output = 0.0;
    a->Ninp = N;
    a->input = 0.0;
    a->output_func = output_func;
    a->old_output = 0.0;
    /*
    a->self_wgt = 0.0;
    */
    a->state = 0.0;
    if ( a->Ninp == 0 ) {
        a->inp_wgt = NULL;
        a->inp_neuron = NULL;
        return;
    }

    a->inp_wgt = (double *)malloc(sizeof(double) * (a->Ninp+1));
    if ( a->inp_wgt == NULL){
        fprintf(stderr, "Can not allocate memory\n");
        exit (EXIT_FAILURE);
    }
    for (i=0; i<= a->Ninp; i++) {
        a->inp_wgt[i] = MTgenrand_f() * range;
    }

    a->inp_neuron = (neuron **)malloc(sizeof(neuron *) * (a->Ninp+1));
    if ( a->inp_neuron == NULL ) {
        fprintf(stderr, "Can not allocate memory\n");
        exit (EXIT_FAILURE);
    }
    for (i=0; i <= a->Ninp; i++) {
        a->inp_neuron[i] = NULL;
    }
}


int ParseArgs(int argc, char **argv)
/* Parse command line arguments and set the variables */
/* The variables to be set are 
   learning_ratio, seed, error_limit,
   EPOCH, N_of_hidden_layer, TO, 
   input_file_name, init_file_name */
{
    int argn=1;

    while ( argn < argc ) {
        if ((strcmp(argv[argn],"--help") == 0) || (strcmp(argv[argn],"-h") == 0)) {
            usage(argv[0]);     exit (EXIT_FAILURE);
        }
	if (strcmp(argv[argn], "-verbose") == 0) {
	    verbose = True;
	}
	if (strcmp(argv[argn], "-very_verbose") == 0) {
	    very_verbose = True;
                verbose = True;
	}
        if ((argv[argn][0] == '-') && (argv[argn+1] != NULL)) {
            if (strcmp(argv[argn], "-k") == 0) {
                N_of_winners = atoi(argv[++argn]);
		if ( N_of_winners <= 0 ) {
		    usage(argv[0]);  
		    fprintf(stderr, "### !!!!! Invalid -k parameter !!!! \n");
		    exit (EXIT_FAILURE);
		}
            }
            if (strcmp(argv[argn], "-leakage") == 0) {
                leakage = atof(argv[++argn]);
            }
            if (strcmp(argv[argn], "-k_bias") == 0) {
                k_bias = atof(argv[++argn]);
            }
            if (strcmp(argv[argn], "-n2k_limit") == 0) {
                n2k_limit = atof(argv[++argn]);
            }
            if (strcmp(argv[argn], "-n2k_threshold") == 0) {
                n2k_threshold = atof(argv[++argn]);
            }
            if (strcmp(argv[argn], "-learn") == 0) {
                learning_ratio = atof(argv[++argn]);
            }
            if (strcmp(argv[argn], "-rand_range") == 0) {
                rand_range = atof(argv[++argn]);
            }
            if (strcmp(argv[argn],"-seed") == 0) {
                seed = atoi(argv[++argn]);
            }
            if (strcmp(argv[argn], "-error") == 0) {
                error_limit = atof(argv[++argn]);
            }
            if (strcmp(argv[argn], "-epoch") == 0) {
                EPOCH = atoi(argv[++argn]);
            }
            if (strcmp(argv[argn], "-hidden") == 0) {
                N_of_hidden_layer = atoi(argv[++argn]);
            }
            if (strcmp(argv[argn], "-to") == 0) {
                TO = atoi(argv[++argn]);
		if ( TO < 0 ) {
		    usage(argv[0]);  
		    fprintf(stderr, "### !!!!! Invalid -to parameter !!!! \n");
		    exit (EXIT_FAILURE);
		}
            }
            if (strcmp(argv[argn], "-input") == 0) {
		input_file_name = (char *)malloc(strlen(argv[++argn])+1);
		if (input_file_name == NULL) {
		    fprintf(stderr, "Could not allocate memory in main()\n");
		    exit (EXIT_FAILURE);
		}
		if (strcpy(input_file_name, argv[argn]) == NULL) {
		    fprintf(stderr, "Could not copy filename in main()\n");
		    exit (EXIT_FAILURE);
		}
		input_file_name[strlen(argv[argn])] = '\0';
	    }
            if (strcmp(argv[argn], "-init") == 0) {
		init_file_name = (char *)malloc(strlen(argv[++argn])+1);
		if (init_file_name == NULL) {
		    fprintf(stderr, "Could not allocate memory in main()\n");
		    exit (EXIT_FAILURE);
		}
		if (strcpy(init_file_name, argv[argn]) == NULL) {
		    fprintf(stderr, "Could not copy filename in main()\n");
		    exit (EXIT_FAILURE);
		}
		init_file_name[strlen(argv[argn])] = '\0';
	    }
	}
	argn++;
    }

    if ( N_of_hidden_layer == 0 ) {
	usage(argv[0]);
	fprintf(stderr, "### !!! Invalid number of hidden layer(%d) !!!\n", N_of_hidden_layer);
	exit (EXIT_FAILURE);
    }
    if ( N_of_winners > N_of_hidden_layer ) {
	usage(argv[0]);
	fprintf(stderr, "### !!! Invalid -k %d !!!\n", N_of_winners);
	exit (EXIT_FAILURE);
    }	
    if (input_file_name == NULL) {
	usage(argv[0]);
	fprintf(stderr, "### !!! Invalid filename of input=%s !!!\n", input_file_name);
	exit (EXIT_FAILURE);
    }
    Input_Info = matinfo_read(input_file_name);

    N_of_input_layer = Input_Info->m;
    if ( N_of_input_layer <= 0 ) { usage(argv[0]); exit (EXIT_FAILURE);}

    if ( N_of_hidden_layer == 0 ) { usage(argv[0]); exit (EXIT_FAILURE);}

    NPATS = Input_Info->n;
    example = Input_Info->mat;

    return OK;
}

void Initialize_network(void) 
{
    int i, j;
    FILE *fp;
    matrix TMP;

    input = (neuron *)malloc(sizeof(neuron) * (N_of_input_layer+1));
    if ( input == NULL ) {
	fprintf(stderr, "Could not allocate memroy in Initialize_network()\n");
	exit (EXIT_FAILURE);
    }
    hidden = (neuron *)malloc(sizeof(neuron) * (N_of_hidden_layer+1));
    if ( hidden == NULL ) {
	fprintf(stderr, "Could not allocate memroy in Initialize_network()\n");
	exit (EXIT_FAILURE);
    }

    /* initialize each neuron in hidden layer */
    /* hidden[0] will be treated as a threshold in the output layer */
    /*
    initialize_neuron(&hidden[0],0, linear_f, rand_range);  
     */
    initialize_neuron(&hidden[0],0, tanh_f, rand_range);  
    hidden[0].output = 1.0; 
    for (i=1; i<=N_of_hidden_layer; i++) { 
        initialize_neuron(&hidden[i], N_of_input_layer, tanh_f, rand_range); 
    }
    
    /* initialize each neuron in input layer */
    /* input[0] will be treated as a threshold in the hidden layer */
    initialize_neuron(&input[0],0, linear_f, rand_range);   
    input[0].output = 1.0;
    for (i=1; i<=N_of_input_layer; i++) { 
        initialize_neuron(&input[i],0, linear_f, rand_range); 
    }
    /* NOTE: */    
    /* The output values both input[0] and hidden[0] are              */
    /* dealed as threshold values, therefore, the output of each      */
    /* should be constantly set to 1.                                 */
    
    /********* Making connection among neurons **********/
    /* From input layer to hidden layer */
    for (i=1; i<=N_of_hidden_layer; i++) {  
        for (j=0; j<=N_of_input_layer; j++) {
            hidden[i].inp_neuron[j] = &input[j]; 
        }
    }

    /* When init_file_name will be given, then          */
    /* read the initial value of weight value from file */
    if ( init_file_name != NULL ) {

	TMP = matinit(1,N_of_input_layer);
	fp = fileio_util_openr(init_file_name);

	for (i=1; i<=N_of_hidden_layer; i++) {  
	    freadmat(fp, TMP, 1, N_of_input_layer);
	    for (j=1; j<=N_of_input_layer; j++) {
		hidden[i].inp_wgt[j] = *(TMP+offs(1,N_of_input_layer,1,j));
	    }
        }
	free(TMP);
	fileio_util_close(fp);
    }
}

/* Set values for an input pattern to  each input neron */
/* where pattern number will be given by second argument pat(int) */
void set_input(neuron *input, int pat) 
{
    int i;
    
    input[0].output = 1.0; /* threshold */
    for (i=1; i<=N_of_input_layer; i++) {
	input[i].output = *(example+offs(NPATS,N_of_input_layer,pat+1,i));
    }
}

void print_neuron(neuron *a)
{
    int i;

    fprintf(stdout, "### output :%5.3f  ", a->output);
    fprintf(stdout, "### Ninp   :%d\n", a->Ninp);
    for (i=1; i<= a->Ninp; i++) {
	fprintf(stdout, "%6.3f ", a->inp_wgt[i]);
/*
	fprintf(stdout, "[%02d]=%6.3f ", i, a->inp_wgt[i]);
	if ( (i % 5 )== 0 ) {
	    fprintf(stdout, "\n");
	}
*/
    }
    fprintf(stdout, "\n");
}


void calc_output (neuron *a) 
{
    double wrk = 0.0;
    int i;

    for(i=1; i<= a->Ninp; i++){
	wrk += a->inp_wgt[i] * a->inp_neuron[i]->output;
    }
    a->input = wrk;
    a->state = a->input;
    a->output = a->output_func(a->state);
}

/* Update weights by cpca rule */
double update_weight(neuron *a)
{
    int i;
    double dwt, ret;

    dwt = 0.0;
    for (i=0; i<= a->Ninp; i++) {
	dwt = a->output * ( a->inp_neuron[i]->output - a->inp_wgt[i] );
	if ( a->output > 0 ) {
	    a->inp_wgt[i] += learning_ratio * dwt;
	}
    }
    ret = (a->output - a->old_output) * (a->output - a->old_output);
    return ret;
}

/*
double diff_weight(neuron *a)
{
    int i;
    double dwt;

    dwt = 0.0;
    for (i=0; i<= a->Ninp; i++) {
	dwt += (a->inp_wgt[i] - a->old_inp_wgt[i]) * (a->inp_wgt[i] - a->old_inp_wgt[i]);
	a->old_inp_wgt[i] =  a->inp_wgt[i];
    }
    return dwt;
}
*/

void do_shuffle(int *order,int n)
{
    int i, index, tmp;

    for (i=0; i<n; i++) {
        index = i + (int)(MTgenrand_f() * ((double)n - (double)i ));
	tmp = order[index];
	order[index] = order[i];
	order[i] = tmp;
    }
    /*
    for (i=0; i<n; i++) {
	fprintf(stderr, "### [%d]=%d\n",i,order[i]);
    }
    */
}

/* k-winners-take-all circuit phase */
void kWTA()
{
    double dwt, n2k, old_n2k;
    int count_k, i, t;

    leakage = LEAKAGE;
    n2k = 0.0;
    /*
    if (verbose) {
	fprintf(stderr, "############ begining of kWTA ############\n");
    }
    */
    t = 0;
    do {
	old_n2k = n2k;
	n2k = 0.0;
	for (i=1; i<=N_of_hidden_layer; i++) {
	    n2k += hidden[i].output;
	}
	n2k += (double)N_of_hidden_layer -2.0 * (double)(N_of_winners);
	
	dwt = 0.0;
	count_k = 0;
	for (i=1; i<=N_of_hidden_layer; i++) {
	    hidden[i].state += hidden[i].output_func(hidden[i].state)
		- leakage * hidden[i].output - k_bias * n2k;
	    /*
	    hidden[i].old_output = hidden[i].output;
	    hidden[i].state += hidden[i].input 
		- leakage * hidden[i].output
		- k_bias * n2k;
	    */
	    hidden[i].output = hidden[i].output_func(hidden[i].state);
	    dwt += (hidden[i].old_output - hidden[i].output) * (hidden[i].old_output - hidden[i].output);
	    
	    if ( hidden[i].output >= 0.0 ) {
		count_k++;
	    }
	    /*
	      if (verbose) {
	      fprintf(stdout, "[%d]=%5.2f,", i,hidden[i].output);
	      }
	    */
	}
	if (verbose) {
	    fprintf(stderr, 
		    "# t=%d n2k=%g, old_n2k=%g, diff=%g count_k=%d(%d)   \n", 
		    t, n2k, old_n2k, fabs(n2k - old_n2k), count_k, 
		    N_of_winners);
	}
	if ( fabs(n2k - old_n2k) < n2k_limit) {
	    t++;
	}
	if (( t > 100 ) && (fabs(n2k - old_n2k) < n2k_limit)) {
	    printf("#### leakage should be changed to %g####\n", leakage);
	    exit (EXIT_FAILURE);
	}
    } while ((count_k != N_of_winners) || (fabs( n2k - old_n2k ) > n2k_limit));
/*
    } while ((count_k != N_of_winners) && (fabs( n2k - old_n2k ) < n2k_limit));
*/
}

/* main loop of computation */
long int main_loop () 
{
    long int t;
    int i, pat;
    int *shuffle;
    double ssw = 0.0, old_ssw = 0.0;

    shuffle=(int *)malloc(sizeof(int) * NPATS);
    for ( i=0; i<NPATS; i++ ) {
	shuffle[i] = i;
    }

    if (verbose) {
	fprintf(stdout, "### inital weights ###\n");
	for (i=1; i<=N_of_hidden_layer; i++) {
	    print_neuron(&hidden[i]);
	}
    }

    for (t=FROM; t<TO; t++){ /* learning time */
	do_shuffle(shuffle,NPATS);
	old_ssw = ssw;
	ssw = 0.0;
        for (pat=0; pat<NPATS; pat++){ /* foreach pattern */
            set_input(input, shuffle[pat]);
            for (i=1; i<=N_of_hidden_layer; i++) {
                calc_output(&hidden[i]);
            }
	    kWTA();

            /* CPCA learning phase */
            for (i=1; i<=N_of_hidden_layer; i++) {
                ssw += update_weight(&hidden[i]);
		if (very_verbose) {
		    print_neuron(&hidden[i]);
		}
            }
        } /* end of each pattern */

	if ( (t % EPOCH) == 0) {
	    fprintf(stdout, "### iter=%ld, ssw=%g, diff=%g (%g) ", 
		    t, ssw, fabs(ssw - old_ssw),
		    (double)NPATS * (double)N_of_hidden_layer * error_limit);
	    fprintf(stdout, "\n");
	    if (verbose) {
		for (i=1; i<=N_of_hidden_layer; i++) {
		    print_neuron(&hidden[i]);
		}
	    }
	}

	if ( fabs(ssw - old_ssw) < error_limit ){
	    fprintf(stdout, "### iter=%ld !!!CONVERGED!!!\n",t);
	    return t;
	}
    }
    return t;
}

int main(int argc, char **argv) 
{
    long int iterations = 0, i;

    ParseArgs(argc, argv);

    init_random(seed);
    Initialize_network();

    
    iterations = main_loop();

    /* print result */
    fprintf(stdout, "### Final weights matrices ###\n");
    for (i=1; i<=N_of_hidden_layer; i++) {
	print_neuron(&hidden[i]);
    }


    fprintf(stdout, "### %s -hidden %d -input %s -k %d -leakage %g -k_bias %g -n2k_limit %g -n2k_threshold %g %s %s -learn %g -error %g -epoch %d -to %d -seed %d -rand_range %g ---> converged at %ld iterations\n",
	    argv[0], N_of_hidden_layer, input_file_name, 
	    N_of_winners, leakage, k_bias, n2k_limit, n2k_threshold,
            (init_file_name != NULL) ? "-init" : "",
            (init_file_name != NULL) ? init_file_name : "",
	    learning_ratio, 
	    error_limit, EPOCH, TO, 
	    get_seed(), rand_range, iterations);

    return OK;
}
