/*
 *  @(#)fft.c - Fast Fourier Transformation 1.0
 *  References: okumura
 */
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <errno.h>
#include "matrix_util.h"
#include "fileio_util.h"
#include "fft.h"

/* make a sin table, for the sake of fft */
static void make_sintbl(int n, double sintbl[])
{
    int i, n2, n4, n8;
    double c, s, dc, ds, t;
    
    n2 = n / 2;  n4 = n / 4;  n8 = n / 8;
    t = sin(PI / n);
    dc = 2 * t * t;  ds = sqrt(dc * (2 - dc));
    t = 2 * dc;  c = sintbl[n4] = 1;  s = sintbl[0] = 0;
    for (i = 1; i < n8; i++) {
        c -= dc;  dc += t * c;
        s += ds;  ds -= t * s;
        sintbl[i] = s;  
        sintbl[n4 - i] = c;
    }
    if (n8 != 0)
        sintbl[n8] = sqrt(0.5);
    for (i = 0; i < n4; i++)
        sintbl[n2 - i] = sintbl[i];
    for (i = 0; i < n2 + n4; i++)
        sintbl[i + n2] = - sintbl[i];
}

/*  make a bit reverse table, for the sake of fft */
static void make_bitrev(int n, int bitrev[])
{
    int i, j, k, n2;
    
    n2 = n / 2;
    i = j = 0;
    for (;;) {
        bitrev[i] = j;
        if (++i >= n) break;
        k = n2;
        while (k <= j) {
            j -= k;
            k /= 2;
        }
        j += k;
    }
}

/*******************************************************************
 * Fast Fourier Transform (Cooley--Tukey algorithm).
 * Notice: number of sample(N) must be limited 2^N.
 * Re[k] : real part, Im[k] : imaginary part, where k = 0,1,2,...
 * Results would be overwritten on Re[], Im[]
 * If inverse flag(4th argument) will be true, then inverse
 * transformation would be performed.
 * If opt flag(5th argument) will be true, then optical fft,
 * which means that DC component would be centerd, must be performed.
 *******************************************************************/
int fft(int n, matrix Re, matrix Im, int inverse, int opt)
{
    static int    last_n  = 0;
    static int    *bitrev = NULL; /* bit reverse table */
    static double *sintbl = NULL; /* table of triangle functions */
    scalar t, s, c, dx, dy;
    int i, j, k, ik, h, d, k2, n4;

    /* preparation */
    n4 = n / 4;
    if ( n != last_n ){
        last_n = n;
        if (sintbl != NULL) free(sintbl);
        if (bitrev != NULL) free(bitrev);
        sintbl = (double *)malloc((n + n4) * sizeof(double));
        bitrev = (int *)malloc(n * sizeof(int));
        if (sintbl == NULL || bitrev == NULL) {
            perror("Could not allocate memory for sintbl in fft()");
            exit (EXIT_FAILURE);
        }
        make_sintbl(n, sintbl);
        make_bitrev(n, bitrev);
    }
    
    if (opt){
        for (i=1; i<n; i+=2){
            Re[i] = -Re[i]; Im[i] = -Im[i];
        }
    }
    
    /* bit reverse */
    for (i=0; i<n; i++) {
        j = bitrev[i];
        if (i < j) {
            t = Re[i];  Re[i] = Re[j];  Re[j] = t;
            t = Im[i];  Im[i] = Im[j];  Im[j] = t;
        }
    }
    
    /* transform */
    for (k = 1; k < n; k = k2) {
        h = 0;  k2 = k + k;  d = n / k2;
        for (j = 0; j < k; j++) {
            c = sintbl[h + n4];
            if (inverse == -1) s = - sintbl[h];
            else               s =   sintbl[h];
            for (i = j; i < n; i += k2) {
                ik = i + k;
                dx = s * Im[ik] + c * Re[ik];
                dy = c * Im[ik] - s * Re[ik];
                Re[ik] = Re[i] - dx;  Re[i] += dx;
                Im[ik] = Im[i] - dy;  Im[i] += dy;
            }
            h += d;
        }
    }
    
    if (opt){
        for (i=1; i<n; i+=2) {
            Re[i] = -Re[i]; Im[i] = -Im[i];
        }
    }
    
    /* if not inverse transformation, then divide n */
    if (!(inverse == -1)){
        for (i = 0; i < n; i++){
            Re[i] /= n;  Im[i] /= n;
        }
    }
    return OK;
}

/*******************************************************************
 * Notice: fft() should be required.
 * Re[k*k] : real part, Im[k*k] : imaginary part, 
 * where k = 0,1,2,...
 * Results would be overwritten on Re[], Im[]
 * If inverse flag(4th argument) will be true, then inverse
 * transformation would be performed.
 *******************************************************************/
int fft2d(int n, matrix Re, matrix Im, int inverse)
{
    matrix Re1, Im1;
    int    i,j;
    
    Re1 = matinit(n,1);
    Im1 = matinit(n,1);
    
    for ( i=1; i<=n; i++){
        for ( j=1; j<=n; j++){
            *(Re1+offs(n,1,j,1)) = *(Re+offs(n,n,j,i));
            *(Im1+offs(n,1,j,1)) = *(Im+offs(n,n,j,i));
        }
        if (!fft(n, Re1, Im1, inverse, 1))
            return EXIT_FAILURE;
        for ( j=1; j<=n; j++){
            *(Re+offs(n,n,j,i)) = *(Re1+offs(n,1,j,1));
            *(Im+offs(n,n,j,i)) = *(Im1+offs(n,1,j,1));
        }
    }
    for ( i=1; i<=n; i++){
        for ( j=1; j<=n; j++){
            *(Re1+offs(n,1,j,1)) = *(Re+offs(n,n,i,j));
            *(Im1+offs(n,1,j,1)) = *(Im+offs(n,n,i,j));
        }
        if (!fft(n, Re1, Im1, inverse, 1))
            return EXIT_FAILURE;
        for ( j=1; j<=n; j++){
            *(Re+offs(n,n,i,j)) = *(Re1+offs(n,1,j,1));
            *(Im+offs(n,n,i,j)) = *(Im1+offs(n,1,j,1));
        }
    }
    free (Re1); free (Im1);
    return OK;
}

#ifdef STANDALONE
#include <stdio.h>

void usage(char *prgname)
{
    fprintf(stderr, "%s - Fast Fourier Transform by Cooley-Tukey's Algorithm\n", prgname);
    fprintf(stderr, "Usage: %s [-i] [datafile]\n", prgname);
    fprintf(stderr, "Option -i: inverse fft\n");
    fprintf(stderr, "       --help: show this message\n");
    exit (EXIT_FAILURE);
}

int main(int argc, char **argv)
{
    matinfo In;
    matrix Re, Im;
    int argn=1, size = 0;
    char *filename;
    FLAG Inverse = False;
    
    while(argn < argc && argv[argn][0] == '-' && argv[argn][1] != '\0'){
        if (keymatch(argv[argn], "--help")) {
            usage(argv[0]);
        } else if (keymatch(argv[argn], "-h")) {
            usage(argv[0]);
        } else if (keymatch(argv[argn], "-i")) {
            Inverse = True;
        } else { usage(argv[0]); }
        argn++;
    }

    if ( argn != argc ) { 
        if ((filename=(char *)malloc(strlen(argv[argn])+1)) == NULL) {
            fprintf(stderr, "Could not allocate memory at fft() - %s", strerror(errno));
            exit (EXIT_FAILURE);
        }
        strcpy(filename, argv[argn]);
        *(filename+strlen(argv[argn])) = '\0';
    } else { 
        if ((filename = (char *)malloc(2)) == NULL) {
            fprintf(stderr, "Could not allocate memory at fft() - %s", strerror(errno));
            exit (EXIT_FAILURE);
        }
        filename[0]='-'; filename[1]='\0';
    }

    In = matinfo_read(filename);

    if (Inverse) {
        size = In->n >> 1; 
    } else { 
        size = In->n; 
    }
    if ( size == 0 ) {
        fprintf(stderr, "Illegal size(%d) of data\n", size);
        exit (EXIT_FAILURE);
    }
    
    Re = matTrim(In->mat, In->n, In->m, 1, 1, size, In->m);
    if (!Inverse) {
	Im = matinit(size,In->m);
    } else {
	Im = matTrim(In->mat, In->n, In->m, size+1, 1, In->n, In->m);
    }
    
    fprintf(stdout, "### Progname: %s %s %s\n", argv[0], 
        Inverse ? "-i":"", filename);
    fprint_signature(stdout);

    if (!Inverse) {
        if (!fft(size, Re, Im, 1, 1))
            return EXIT_FAILURE;
        fwritemat(stdout, Re, size, 1, "### Real Part:\n", "");
        fwritemat(stdout, Im, size, 1, "### Imaginary Part:\n", "");
    } else {
        if (!fft(size, Re, Im, -1, 1)) 
            return EXIT_FAILURE;
        fwritemat(stdout, Re, size, 1, "### Result of Inverse FFT:\n", "");
    }
    return EXIT_SUCCESS;
}
#endif STANDALONE
