/* 
 * A k-Winners-take-all circuit simulator
 *
 * ProgramName: kWTA.c
 * Author: Shinichi ASAKAWA <asakawa@twcu.ac.jp>
 * $Id: kWTA.c,v 1.2 2001/11/24 09:48:27 asakawa Exp asakawa $
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include "general.h"
#include "fileio_util.h"
#include "matrix_util.h"

double leakage = 0.01;
double k_bias = 1.0;
double n2k_limit = 0.001;

struct NEURON {
    double output;              /* Output values of this neuron */
    double (*output_func)();    /* the pointer to the output function */
    double input;
    double state;
    double old_output;
    double self_wgt;
};
typedef struct NEURON neuron;


/* ------------------------------------------------------------------------
   initial set of global variables
   ------------------------------------------------------------------------ */
int FROM=0;
int TO=10000;
int seed=0;

/* NOTE: 0-th neuron will be handled as a threshold value
   The actual number of neurons in each layer will be automatically
   settled by program in accordance with both the command line arguments
   and input file. */
int N_of_hidden_layer = 0;
int N_of_winners = 1;

neuron *hidden;

int NPATS = 0;               /* Number of patterns to be learned */
char *input_file_name = NULL;
matrix example;
matinfo Input_Info;

/* initialize a neuron which has N inputs */
void initialize_neuron(neuron *a, double (*output_func)());

void calc_output(neuron *a);

void usage(char *prgname)
{
    fprintf(stderr, "### Program name: %s - A k-Winners-Take-All circuit simulator\n", prgname);
    fprintf(stderr, "### Creator: Shinichi ASAKAWA <asakawa@twcu.ac.jp>\n");
    fprintf(stderr, "Usage: %s -input <filename> [-k <int>] [-leakage <double>] [-k_bias <double>] [-to <int>]\n", prgname);
    fprintf(stderr, "Arguments: \n");
    fprintf(stderr, "     input: input file name to be learned\n");
    fprintf(stderr, "default settings of options: \n");
    fprintf(stderr, "      k = %d number of winners\n", N_of_winners);
    fprintf(stderr, "      leakage = %g\n", leakage);
    fprintf(stderr, "      k_bias = %g\n", k_bias);
    fprintf(stderr, "      to = %d: max iterations\n", TO);
}

double tanh_f(double state)
{
    return tanh(state);
}

double sigmoid_f(double state)
{
    return 1.0/(1.0 + exp(-4.0 * state));
}

/* Initilazing a neuron, which means */
/* assigning an initial value as default to each variable  */
void initialize_neuron(neuron *a, double output_func())
{
    a->output = 0.0;
    a->output_func = output_func;
    a->input = 0.0;
    a->state = 0.0;
    a->old_output = 0.0;
    a->self_wgt = leakage;
}

int ParseArgs(int argc, char **argv)
/* Parse command line arguments and set the variables */
/* The variables to be set are 
   TO, leakage, k_bias, input_file_name, n2k_limit */
{
    int argn=1;

    while ( argn < argc ) {
        if ((strcmp(argv[argn],"--help") == 0) || (strcmp(argv[argn],"-h") == 0)) {
            usage(argv[0]);     exit (EXIT_FAILURE);
        }
        if ((argv[argn][0] == '-') && (argv[argn+1] != NULL)) {
            if (strcmp(argv[argn], "-to") == 0) {
                TO = atoi(argv[++argn]);
		if ( TO < 0 ) {
		    usage(argv[0]);  
		    fprintf(stderr, "### !!!!! Invalid -to parameter !!!! \n");
		    exit (EXIT_FAILURE);
		}
            }
            if (strcmp(argv[argn], "-k") == 0) {
                N_of_winners = atoi(argv[++argn]);
		if ( N_of_winners <= 0 ) {
		    usage(argv[0]);  
		    fprintf(stderr, "### !!!!! Invalid -k parameter !!!! \n");
		    exit (EXIT_FAILURE);
		}
            }
            if (strcmp(argv[argn], "-leakage") == 0) {
                leakage = atof(argv[++argn]);
            }
            if (strcmp(argv[argn], "-k_bias") == 0) {
                k_bias = atof(argv[++argn]);
            }
            if (strcmp(argv[argn], "-n2k_limit") == 0) {
                n2k_limit = atof(argv[++argn]);
            }
            if (strcmp(argv[argn], "-input") == 0) {
		input_file_name = (char *)malloc(strlen(argv[++argn])+1);
		if (input_file_name == NULL) {
		    fprintf(stderr, "Could not allocate memory in main()\n");
		    exit (EXIT_FAILURE);
		}
		if (strcpy(input_file_name, argv[argn]) == NULL) {
		    fprintf(stderr, "Could not copy filename in main()\n");
		    exit (EXIT_FAILURE);
		}
		input_file_name[strlen(argv[argn])] = '\0';
	    }
	}
	argn++;
    }

    if (input_file_name == NULL) {
	usage(argv[0]);
	fprintf(stderr, "### !!! Invalid filename of input=%s !!!\n", input_file_name);
	exit (EXIT_FAILURE);
    }
    Input_Info = matinfo_read(input_file_name);

    N_of_hidden_layer = Input_Info->m;
    if ( N_of_hidden_layer <= 0 ) { usage(argv[0]); exit (EXIT_FAILURE);}
    NPATS = Input_Info->n;
    example = Input_Info->mat;

    if (N_of_winners > N_of_hidden_layer ) {
	usage(argv[0]);
	fprintf(stderr, "### !!! Invalid -k=%d !!!\n", N_of_winners);
	exit (EXIT_FAILURE);
    }
    return OK;
}

void Initialize_network(void) 
{
    int i;

    hidden = (neuron *)malloc(sizeof(neuron) * (N_of_hidden_layer+1));
    if ( hidden == NULL ) {
	fprintf(stderr, "Could not allocate memroy in Initialize_network()\n");
	exit (EXIT_FAILURE);
    }
    /* initialize each neuron in hidden layer */
    /* hidden[0] will be treated as a threshold in the hidden layer */
    initialize_neuron(&hidden[0], tanh_f);
    hidden[0].output = 1.0;
    for (i=1; i<=N_of_hidden_layer; i++) { 
	initialize_neuron(&hidden[i], tanh_f);
	/* initialize_neuron(&hidden[i], sigmoid_f);*/
    }
}


/* main loop of computation */
int main_loop () 
{
    int t,i, pat, count_k;
    double n2k, old_n2k, dwt, mdwt;

    /* Set values for an input pattern to  each input neron */
    /* where pattern number will be given by second argument pat(int) */
    for (pat=0; pat<NPATS; pat++){

	for (i=1; i<=N_of_hidden_layer; i++) {
	    hidden[i].input = *(example+offs(NPATS,N_of_hidden_layer,pat+1,i));

	    hidden[i].state = hidden[i].input;
	    hidden[i].output = hidden[i].output_func(hidden[i].state);
	    hidden[i].old_output = 0.0;
	}

	fprintf(stdout, "### Initial value (input) ");
	for (i=1; i<=N_of_hidden_layer; i++) {
	    fprintf(stdout, "%d=%g, ",i,hidden[i].input);
	}
	fprintf(stdout,"\n");

	fprintf(stdout, "### Initial value (state) ");
	for (i=1; i<=N_of_hidden_layer; i++) {
	    fprintf(stdout, "%d=%g, ",i,hidden[i].state);
	}
	fprintf(stdout,"\n");

	fprintf(stdout, "### Initial value (output) ");
	for (i=1; i<=N_of_hidden_layer; i++) {
	    fprintf(stdout, "%d=%g, ", i,hidden[i].output);
	}
	fprintf(stdout,"\n");

	/*
	for (t=FROM; t<TO; t++){
	*/
	t=0;
	do {
	    t++;
	    fprintf(stderr, "### t=%d, ", t);

	    old_n2k = n2k;
	    n2k = 0.0;
            for (i=1; i<=N_of_hidden_layer; i++) {
		n2k += hidden[i].output;
	    }
	    n2k += (double)N_of_hidden_layer - 2.0 * (double)(N_of_winners);
	    fprintf(stderr, "n2k=%6.3f ", n2k);

	    dwt = 0.0;
	    count_k=0;
            for (i=1; i<=N_of_hidden_layer; i++) {
		hidden[i].state += hidden[i].output_func(hidden[i].input)
		               - leakage * hidden[i].output
                               - k_bias * n2k;
		hidden[i].output = hidden[i].output_func(hidden[i].state);
            if (hidden[i].output > 0.0) {
		count_k++;
	    }
 /*
		hidden[i].state += hidden[i].input 
		               - leakage * hidden[i].output
                               - k_bias * n2k;
		hidden[i].output = hidden[i].output_func(hidden[i].state);
*/
		dwt += (hidden[i].old_output - hidden[i].output) * (hidden[i].old_output - hidden[i].output);

		hidden[i].old_output = hidden[i].output;
		fprintf(stderr, "[%d]=%4.2f,", i,hidden[i].output);
	    }
	    mdwt = dwt / (double)(N_of_hidden_layer);
	    fprintf(stderr, ":count_k=%d\n", count_k);
            /*
	    fprintf(stderr, ":mdwt=%g\n", mdwt);
	    if (mdwt < n2k_limit ) { return t;}
	    */
	    if ((fabs( n2k - old_n2k ) < n2k_limit) && (count_k < N_of_winners) && (count_k != 0 )) {
		leakage *= 2.0;
		fprintf(stderr, "#### difference too small leakage=%g\n",leakage);
/*
		return 0;
*/
             }
	} while ( count_k != N_of_winners );
/*
	} while ( (count_k == 0 ) || (count_k > N_of_winners ));
	} while ((fabs( n2k - old_n2k ) > n2k_limit) || (fabs(n2k)> 1.0));
*/
	    /*
	    if ((fabs( n2k - old_n2k ) < n2k_limit) && ( fabs(n2k) < 1.0)) {
		return t;
	    }
	    */
    }
    return t;
}

int main(int argc, char **argv) 
{
    long int iterations = 0;

    ParseArgs(argc, argv);
    Initialize_network();
    iterations = main_loop();

    fprintf(stdout, "### %s -input %s -to %d -k %d -leakage %g -k_bias %g -n2k_limit %g\n",
	    argv[0], input_file_name, 
	    TO, N_of_winners, leakage, k_bias, n2k_limit);
    return OK;
}
