/*  Young-Householder transform, double centering 

Copyright (c) 1999 Shinichi ASAKAWA <asakawa@twcu.ac.jp>

Disclaimer:  No guarantees of performance accompany this software,
nor is any responsibility assumed on the author.  */

#include <stdlib.h>
#include <math.h>
#include "matrix_util.h"

int matDC(matrix a, int n, int m)
{
    vector temp;
    int size = ((n>=m) ? n : m);
    int i,j;
    scalar a1, a2;
    
    temp = vectinit(size);
    for (a1=0.0, i=1; i<=n; i++) {
	a2 = 0.0;
	for (j=1; j<=n; j++)
	    a2 += *(a+offs(n,m,i,j));
	a1 += a2;
	*(temp+offs(size,1,i,1)) = a2 / (double)n;
    }
    a1 /= (double)n * (double)n;
    
    for (i=1; i<=n; i++) {
	for (j=1; j<=n; j++) {
	    *(a+offs(n,m,i,j)) = 
		*(temp+offs(size,1,i,1)) + *(temp+offs(size,1,j,1))
		- a1 - *(a+offs(n,m,i,j));
	}
    } 
    free(temp); 
    return OK; 
}

#ifdef STANDALONE
#include <stdlib.h>
#include "fileio_util.h"

int usage(char *prgname)
{
    fprintf(stderr,"%s\n", prgname);
    fprintf(stderr,"Usage %s [options] [datafile]\n", prgname);
    fprintf(stderr,"options: \n");
    fprintf(stderr," -p dimensions : size of result dimensions(default=2)\n");
    fprintf(stderr," -g : display on charcter display\n");
    fprintf(stderr," -d : data is distance matrix(default raw data)\n");
    exit (EXIT_FAILURE);
}

/* 
   <-- m -->
   |       |
   n       |  n  m Υǡ󤫤 m  m εΥä
   |       |  metricMDS ¹Ԥ롣
   +-------+
   
   뤤ϡ, -d option ΤȤ m  m  εΥ(гǤ
   ٤ 0)Ϥ metricMDS ¹Ԥ
*/
int main(int argc, char **argv)
{
    int 	i,j,k, n=0, m=0, p=2, argn = 1;
    matinfo In;
    matrix Dist, Raw, evec, result;
    vector eig;
    scalar w, eigchk;
    FLAG plotflg = False, distdataflg = False;
    
    while (argn < argc && argv[argn][0] == '-' && argv[argn][1] != '\0') {
	if (keymatch(argv[argn], "-help")) { usage(argv[0]); } 
	else if (keymatch(argv[argn], "-p")) {
	    ++argn;
	    if (argn==argc || sscanf(argv[argn],"%d",&p) != 1) usage(argv[0]);
	} 
	else if (keymatch(argv[argn], "-d")) { distdataflg = True;  } 
	else if (keymatch(argv[argn], "-g")) { plotflg = True;  } 
	else { usage(argv[0]); }
	++argn;
    }
    
    if (argn != argc) {	In = matinfo_read(argv[argn]);  ++argn; } 
    else              { In = matinfo_read("-"); }
    
    if (argn != argc)  usage(argv[0]);
    
    if (distdataflg) {
	Dist = In->mat;
	m = In->m;
	n = In->n;
	if (n != m) {
	    fileio_util_perror("matDC(main): data file must be a squared matrix");
	    exit (EXIT_FAILURE);
	}
	for (i=1; i<=m; i++) {
	    for (j=1; j<=m; j++) {
		*(Dist+offs(m,m,i,j)) = *(Dist+offs(m,m,i,j)) * *(Dist+offs(m,m,i,j));
            }
        }
    } else {
	Raw = In->mat;
	n = In->n;
	m = In->m;
	Dist = matinit(m,m);
	for (i=1; i<=m; i++) {
	    for (j=1; j<=m; j++) {
		for (k=1; k<=n; k++) {
		    w = *(Raw+offs(n,m,k,i)) - *(Raw+offs(n,m,k,j));
		    *(Dist+offs(m,m,i,j)) += w * w;
		}
	    }
	}
    }
    matARITH(Dist,m,m,0.5,0.0);
    matDC(Dist,m,m);
    
    eig = vectinit(m);
    evec = matinit(m,m);
    result = matinit(m,p);
    eigen_pm(Dist,eig, evec, m,m);
    
    /* check whether eigenvalues are valid or not */
    eigchk = *(eig+offs(m,1,1,1));
    for (i=2;i<=m;i++) {
	if ((eigchk <= *(eig+offs(m,1,i,1))) || (*(eig+offs(m,1,i,1))<eps) || isnan(*(eig+offs(m,1,i,1))) )
	    *(eig+offs(m,1,i,1)) = 0.0;
	eigchk = *(eig+offs(m,1,i,1));
    }
    
    /* check whether eigenvectors are valid or not */
    for (i=1; i<=m; i++) {
	if ( *(eig+offs(m,1,i,1)) == 0.0)
	    for (j=1; j<=m; j++)  *(evec+offs(m,m,j,i)) = 0.0;
    }
    
    for (i=1; i<=m; i++) {
	for (j=1; j<=p; j++){
	    *(result+offs(m,p,i,j)) 
		= *(evec+offs(m,m,i,j)) * sqrt(fabs(*(eig+offs(m,1,j,1))));
	}
    }
    fprintf(stdout, "### ProgramName: %s %s (metric Multi-Dimensional Scaling)\n", argv[0], (argn != argc) ? argv[argn] : "-");
    fprint_signature(stdout);
    fwritemat(stdout, eig,    1,m, "### EigenValues:\n","%.6g ");
    fwritemat(stdout, evec,   m,m, "### EigenVectors:\n", "%.6g ");
    fwritemat(stdout, result, m,p, "### Coodinates:\n", "%.6g ");
    
    if (plotflg) 
	plot2d(stdout, result, m,p, "### Result of MDS");
    return OK;
}
#endif STANDALONE
