/* Compute a correlation matrix 

   Copyright (c) 1999 Shinichi ASAKAWA <asakawa@twcu.ac.jp>

   Disclaimer:  No guarantees of performance accompany this software,
   nor is any responsibility assumed on the author.  */
#include <stdlib.h>
#include <math.h>
#include "matrix_util.h"
#include "fileio_util.h"

int matcorr (matrix A, int n, int m, matrix Corr)
{
    int i, j, k;
    scalar wj, wk;
    matrix mean, sd;

    if (Corr == NULL) {
	fileio_util_perror("matcorr: invalid matrix.");
	exit (EXIT_FAILURE);
    }

    if (A == NULL) {
	fileio_util_perror("matcorr: invalid matrix.");
	exit (EXIT_FAILURE);
    }
    
    if ( (mean = matinit(1,m)) == NULL ) {
	fileio_util_perror("matcorr: can not alloc memory for mean.");
	exit (EXIT_FAILURE);
    }
    if ( (sd = matinit(1,m)) == NULL ) {
	fileio_util_perror("matCorr: can not alloc memory for sd.");
	exit (EXIT_FAILURE);
    }
    
    for (j=1; j<=m; j++)
	for (i=1; i<=n; i++)
	    *(mean+offs(1,m,1,j)) += *(A+offs(n,m,i,j))/(double)n;
    
    for (j=1; j<=m; j++) {
	for (k=1; k<=m; k++) {
	    for (i=1; i<=n; i++) {
		wj = *(A+offs(n,m,i,j)) - *(mean+offs(1,m,1,j));
		wk = *(A+offs(n,m,i,k)) - *(mean+offs(1,m,1,k));
		*(Corr+offs(m,m,j,k)) += wj * wk / (double)n;
	    }
	}
    }
    
    for (j=1; j<=m; j++)
	*(sd+offs(1,m,1,j)) = sqrt(*(Corr+offs(m,m,j,j)));
    
    for (j=1; j<=m; j++)
	for (k=1; k<=m; k++)
	    if ((fabs(*(sd+offs(1,m,1,j))) > eps) && (fabs(*(sd+offs(1,m,1,k))) > eps))
		*(Corr+offs(m,m,j,k)) /= (*(sd+offs(1,m,1,j)) * *(sd+offs(1,m,1,k)));
	    else
		*(Corr+offs(m,m,j,k)) = 0.0;
    
    free(mean); free(sd);
    return OK;
}

matrix matCorr (matrix A, int n, int m)
{
    matrix Ret = matinit(m,m);
    matcorr (A, n, m, Ret);
    return Ret;
}


int matcorrinv(matrix X, int n, int m, scalar *det, matrix corrinv)
{
    matrix ret, ret2;

    if ( X == NULL ) {
	fprintf(stderr, "Invalid matrix  in matcorrinv()\n");
	exit (EXIT_FAILURE);
    }
    if ( corrinv == NULL ) {
	fprintf(stderr, "Could not allocate memory in matcorrinv()\n");
	exit (EXIT_FAILURE);
    }

    ret = matCorr(X,n,m);
    ret2 = matInv(ret,m,det);
    matcpy(ret2,m,m,corrinv);
    free(ret); free(ret2);
    return OK;
}


matrix matCorrinv(matrix X, int n, int m, scalar *det)
{
    matrix ret, ret2;

    if ( X == NULL ) {
	fprintf(stderr, "Invalid matrix  in matCorrinv()\n");
	exit (EXIT_FAILURE);
    }
    
    ret = matCorr(X,n,m);
    ret2 = matInv(ret,m,det);
    free(ret);
    return ret2;
}

int matubcov(matrix X, int n, int m, matrix Ret)
{
    matrix Wrk;

    if ( X == NULL ) {
	fprintf(stderr, "invalid matrix in matUbcov()\n");
	exit (EXIT_FAILURE);
    }
    if ( Ret == NULL ) {
	fprintf(stderr, "Could not allocate memory in matubcov()\n");
	exit (EXIT_FAILURE);
    }
    Wrk = matUbcov(X,n,m);
    matcpy(Wrk,m,m, Ret);
    free (Wrk);
    return OK;
}

/* return unbiased covariance matrix */
matrix matUbcov(matrix X, int n, int m)
{
    int i,j,k;
    scalar wj, wk;
    matrix ret, mean;
    
    if ( X == NULL ) {
	fprintf(stderr, "invalid matrix in matUbcov()\n");
	exit (EXIT_FAILURE);
    }
    mean = bstat_mean(X,n,m);
    ret = matinit(m,m);
    for (j=1; j<=m; j++) {
	for (k=1; k<=m; k++) {
	    for (i=1; i<=n; i++) {
		wj = *(X+offs(n,m,i,j)) - *(mean+offs(1,m,1,j));
		wk = *(X+offs(n,m,i,k)) - *(mean+offs(1,m,1,k));
		*(ret+offs(m,m,j,k)) += wj * wk / (double)(n-1.0);
	    }
	}
    }
    free (mean);
    return ret;
}

/* return sample covariace matrix */
int matcov (matrix X, int n, int m, matrix Cov)
{
    matrix Wrk;

    if ( X == NULL ) {
	fprintf(stderr, "Invalid matrix in matCov()\n");
	exit (EXIT_FAILURE);
    }
    if ( Cov == NULL ) {
	fprintf(stderr, "Could not allocate memory in matcov()\n");
	exit (EXIT_FAILURE);
    }
    Wrk = matCov(X,n,m);
    matcpy(Wrk,m,m,Cov);
    free (Wrk);
    return OK;
}

/* return sample covariace matrix */
matrix matCov (matrix X, int n, int m)
{
    int i, j, k;
    scalar wj, wk;
    matrix ret, mean;

    if ( X == NULL ) {
	fprintf(stderr, "Invalid matrix in matCov()\n");
	exit (EXIT_FAILURE);
    }
    mean = bstat_mean(X,n,m);
    ret = matinit(m,m);
    for (j=1; j<=m; j++) {
	for (k=1; k<=m; k++) {
	    for (i=1; i<=n; i++) {
		wj = *(X+offs(n,m,i,j)) - *(mean+offs(1,m,1,j));
		wk = *(X+offs(n,m,i,k)) - *(mean+offs(1,m,1,k));
		*(ret+offs(m,m,j,k)) += wj * wk / (double)n;
	    }
	}
    }
    free (mean);
    return ret;
}

int matubcovinv (matrix X, int n, int m, scalar *det, matrix Covinv)
{
    matrix ret, ret2;

    if ( X == NULL ) {
	fprintf(stderr, "Invalid matrix in matubcovinv()\n");
	exit (EXIT_FAILURE);
    }
    if ( Covinv == NULL ) {
	fprintf(stderr, "Invalid matrix in matcovinv()\n");
	exit (EXIT_FAILURE);
    }
    
    ret = matUbcov(X,n,m);
    ret2 = matInv(ret,m,det);
    matcpy(ret2,m,m,Covinv);
    free(ret); free(ret2);
    return OK;
}

matrix matUbcovinv (matrix X, int n, int m, scalar *det)
{
    matrix ret, ret2;

    if ( X == NULL ) {
	fprintf(stderr, "Invalid matrix in matUbcovinv()\n");
	exit (EXIT_FAILURE);
    }
	
    ret = matUbcov(X,n,m);
    ret2 = matInv(ret,m,det);
    free(ret);
    return ret2;
}

int matcovinv(matrix X, int n, int m, scalar *det, matrix Covinv)
{
    matrix ret, ret2;
    
    if ( X == NULL ) {
	fprintf(stderr, "Invalid matrix in matUbcovinv()\n");
	exit (EXIT_FAILURE);
    }

    if ( Covinv == NULL ) {
	fprintf(stderr, "Invalid matrix in matcovinv()\n");
	exit (EXIT_FAILURE);
    }
    
    ret = matCov(X,n,m);
    ret2 = matInv(ret,m,det);
    matcpy(ret2,m,m,Covinv);
    free(ret); free(ret2);
    return OK;
}

matrix matCovinv(matrix X, int n, int m, scalar *det)
{
    matrix ret, ret2;
    
    if ( X == NULL ) {
	fprintf(stderr, "Invalid matrix in matCovinv()\n");
	exit (EXIT_FAILURE);
    }
	
    ret = matCov(X,n,m);
    ret2 = matInv(ret,m,det);
    free(ret);
    return ret2;
}

#ifdef STANDALONE
#include <stdlib.h>
#include <stdio.h>

int usage(char *prgname)
{
    printf("%s \n", prgname);
    printf("Usage %s [datafile]\n", prgname);
    exit (EXIT_FAILURE);
}

int main(int argc, char **argv)
{
    int n = 0, m = 0, argn = 1;
    matrix A, Corr = NULL;
    matinfo Ainfo;
    
    while (argn < argc && argv[argn][0] == '-' && argv[argn][1] != '\0') {
        if (keymatch(argv[argn], "--help")) { usage(argv[0]); }
        if (keymatch(argv[argn], "-h")) { usage(argv[0]); }
        ++argn;
    }
    if (argn != argc) {
        Ainfo = matinfo_read( argv[argn] );
    } else {
        Ainfo = matinfo_read("-");
    }
    A = Ainfo->mat;
    n = Ainfo->n; m = Ainfo->m;

    Corr = matCorr(A, n, m);
    fprintf(stdout, "### ProgramName: %s %s\n", argv[0], Ainfo->name);
    fprint_signature(stdout);
    fprintf(stdout, "### DataSize: %d %d\n", m, m);
    fwritemat(stdout, Corr, m, m, "", "");
    return OK;
}
#endif STANDALONE
