/*
 * ProgramName: nmf.c 
 * Author: Shinichi ASAKAWA <asakawa@twcu.ac.jp>
 * $Id: nmf.c,v 1.12 2001/03/31 03:47:29 asakawa Exp asakawa $
 * Date: Mon, 05 Mar. 2001
 * Version: 2.0
 * This version was made for reading data from data file.
 *
 **********************************************************************
 * Original: Daniel D. Lee and Sebastian Seung, (1999), Learning the
 * parts of objects by non-negative matrix factorization, Nature,
 * Vol.401, 788--791.
 **********************************************************************
 *
 * Update history:
 * Date: 12-Dec.-2000
 * Note:
 * This is the my first program for NMF alogrithm.
 * In order to compile this code, you must do in the command line like
 *
 * Note for compilation:
 * gcc -Wall nmf.c -lm
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <unistd.h>
#include "rand_gen.h"
#include "fileio_util.h"
#include "matrix_util.h"
/* Three header files above are nessesary for Asakawa's maniac C libraries */

/* ------------------------------------------------------------------------
   initial set of global variables
   ------------------------------------------------------------------------ */
char *data_file_name=NULL;
char *output_file_name=NULL;
FILE *OUT;

int FROM=0;
int TO=10000;
int EPOCH=20;
int ITER_MAX=10000;
#define DIFF_LIMIT_MIN 0.00000000001
double DIFF_LIMIT=0.0001;
double rand_range = 1.0;
int seed=0;
int verbose=0;

long int N=0; /* N means the number of images */
long int M=0; /* M means the number of pixels */
int R=0;      /* R means the number of basis to be extracted, */
              /*   which means the number of hidden units */

matrix W;     /* weight matrix */
matrix H;     /* hidden units matrix */
matrix V;      /* data matrix */
matinfo V_info;
matrix WH;     /* estimated matrix */
matrix oldWH;

int calcWH(void), updateW(void), updateH(void);
void usage(char *);
int normalize_by_column(matrix, int, int);

int initialize_nmf(void)
{
    int i,j;
    
    V_info=matinfo_read(data_file_name);
    N=V_info->n;
    M=V_info->m;
    
    V = V_info->mat;

    if ( R <= 0 )
	R = ( N > M ) ? M : N;
    
    W = matinit(N, R);
    H = matinit(R, M); 
    WH = matinit(N,M);
    oldWH = matinit(N,M);
    
    /* initialize matrix W by random number */
    for (i=1; i<=N; i++) {
	for (j=1; j<=R; j++) {
	    *(W+offs(N,R,i,j)) = MTgenrand_f() * rand_range;
	}
    }
    normalize_by_column(W,N,R);
    
    /* initialize matrix H by random number */
    for (i=1; i<=R; i++) {
	for (j=1; j<=M; j++) {
	    *(H+offs(R,M,i,j)) = MTgenrand_f() * rand_range;
	}
    }

    return OK;
}

void usage(char *prgname)
{
    fprintf(stderr, "### Program name: %s - A Non-negative Matrix Factarization\n", prgname);
    fprintf(stderr, "### Creator: Shinichi ASAKAWA <asakawa@twcu.ac.jp>\n");
    fprintf(stderr, "Usage: %s -i|--input <filename> -o|--output <output_file_name> [--R <int>] [-h|--help] [--limit <double>] [--epoch <int>] [--rand_range <float> ] [--iter_max <int>] [--verbose] [--seed <int>]\n", prgname);
    fprintf(stderr, "Arguments: \n");
    fprintf(stderr, "     --input: data file name\n");
    fprintf(stderr, "     --output: results file name\n");
    fprintf(stderr, "Options: \n");
    fprintf(stderr, "     --R: number of basis [default=number of lines]\n");
    fprintf(stderr, "     -h|--help: show this help\n");
    fprintf(stderr, "     --limit: limit of difference in iteration [default=%g]\n", DIFF_LIMIT);
    fprintf(stderr, "     --iter_max: number of maximam iterations [default=%d]\n", ITER_MAX);
    fprintf(stderr, "     --rand_range: range of random numbers>0 [default=%g]\n", rand_range);
    fprintf(stderr, "     --output: header of the output files [default stdout]\n");
    fprintf(stderr, "     --epoch: header of the output files [default %d]\n", EPOCH);
    fprintf(stderr, "     --seed: seed for random number generator\n");
    fprintf(stderr, "     --verbose: verbose output\n");
}

int ParseArgs(int argc, char **argv)
{
    int argn=1;
    
    while ( argn < argc ) {
	if ((strcmp(argv[argn],"--help") == 0) || 
	    (strcmp(argv[argn],"-h") == 0)) {
	    usage(argv[0]);     exit (EXIT_FAILURE);
	}
	if ((argv[argn][0] == '-') && (argv[argn+1] != NULL)) {
	    if ((strcmp(argv[argn], "--input") == 0) ||
		strcmp(argv[argn], "-i") == 0) {
		data_file_name = (char *)malloc(strlen(argv[++argn])+1);
		if (data_file_name == NULL) {
		    fprintf(stderr, "Could not allocate memory in ParseArgs()\n");
		    exit (EXIT_FAILURE);
		}
		if (strcpy(data_file_name, argv[argn]) == NULL) {
		    fprintf(stderr, "Could not copy filename in main()\n");
		    exit (EXIT_FAILURE);
		}
		data_file_name[strlen(argv[argn])] = '\0';
	    }
	    if ( (strcmp(argv[argn], "--output") == 0) ||
		 strcmp(argv[argn], "-o") == 0 ) {
		output_file_name = (char *)malloc(sizeof(char) * strlen(argv[++argn])+6+5+4+3);
		if (output_file_name == NULL) {
		    fprintf(stderr, "Could not allocate memory in ParseArgs()\n");
		    exit (EXIT_FAILURE);
		}
		if (sprintf(output_file_name,"%s%c", argv[argn], '\0') <= 0) {
		    fprintf(stderr, "Could not copy output_file_name in ParseArgs()\n");
		    exit (EXIT_FAILURE);
		}
	    }
	    if ((strcmp(argv[argn], "--R") == 0) ||
		strcmp(argv[argn], "-R") == 0) {
		R = atoi(argv[++argn]);
	    }
	    if (strcmp(argv[argn], "--rand_range") == 0) {
		rand_range = atof(argv[++argn]);
	    }
	    if (strcmp(argv[argn], "--limit") == 0) {
		DIFF_LIMIT = atof(argv[++argn]);
	    }
	    if (strcmp(argv[argn], "--iter_max") == 0) {
		ITER_MAX = atoi(argv[++argn]);
	    }
            if (strcmp(argv[argn], "--epoch") == 0) {
                EPOCH = atoi(argv[++argn]);
            }
            if (strcmp(argv[argn],"--seed") == 0) {
                seed = atoi(argv[++argn]);
            }
	    if (strcmp(argv[argn], "--verbose") == 0) {
		verbose=1;
	    }
	}
	argn++;
    }
    
    if (data_file_name == NULL) {
	fprintf(stderr, "Input data file name required\n");
	exit (EXIT_FAILURE);
    }
    if ( ITER_MAX <= 0 ) {
	usage(argv[0]);
	fprintf(stderr, "Invalid --iter_max option %d\n", ITER_MAX);
	exit (EXIT_FAILURE);
    }
    if (rand_range <= 0) {
	usage(argv[0]);
	fprintf(stderr, "Invalid --rand_range %g\n", rand_range);
	exit (EXIT_FAILURE);
    }
    if (DIFF_LIMIT <= DIFF_LIMIT_MIN) {
	usage(argv[0]);
	fprintf(stderr, "Too small limit!\n");
	exit (EXIT_FAILURE);
    }

    if (output_file_name == NULL) {
	output_file_name = (char *)malloc(sizeof(char) * strlen(argv[0])+6+5+4+3);
	if (sprintf(output_file_name,"%s%c", argv[0], '\0') <= 0) {
	    fprintf(stderr, "Could not copy output_file_name in ParseArgs()\n");
	    exit (EXIT_FAILURE);
	}
    }

    return OK;
}

/* compute WH from W and H */
int calcWH(void)
{
    int i,j,k;
    double tmp;
    
    for (i=1; i<=N; i++) {
	for (j=1; j<=M; j++) {
	    tmp = 0.0;
	    for (k=1; k<=R; k++) {
		tmp += *(W+offs(N,R,i,k)) * *(H+offs(R,M,k,j));
#ifdef Isnan
		if (isnan(tmp)) {
		    fprintf(stderr, "NaN at calcWH()\n");
		    exit (EXIT_FAILURE);
		}
#endif
	    }
	    *(WH+offs(N,M,i,j)) = tmp;
#ifdef Isnan
	    if (isnan( *(WH+offs(N,M,i,j)) )) {
		fprintf(stderr, "NaN at calcWH()\n");
		exit (EXIT_FAILURE);
	    }
#endif
	}
    }
    return OK;
}

int normalize_by_column(matrix MAT, int n, int m)
{
    int i,j;
    double tmp;
    
    for (j=1; j<=m; j++) {
	tmp = 0.0;
	for (i=1; i<=n; i++) {
	    tmp += *(MAT+offs(n,m,i,j));
#ifdef Isnan
	    if (isnan(tmp)) {
		fprintf(stderr, "NaN at i%d,j%d,%g in normalize_by_column()\n", i,j, tmp);
		exit (EXIT_FAILURE);
	    }
#endif
	}
	for (i=1; i<=n; i++) {
	    if ( tmp > DIFF_LIMIT ) {
		*(MAT+offs(n,m,i,j)) /= tmp;
#ifdef Isnan
		if ( isnan(*(MAT+offs(n,m,i,j))) ) { 
		    fprintf(stderr, "NaN at i%d,j%d,%g in normalize_by_column()\n", i,j, tmp);
		    exit (EXIT_FAILURE);
		}
#endif
	    }
	}
    }
    return OK;
}

int updateW(void)
{
    int i,a,u;
    double tmp;
    
    for (i=1; i<=N; i++) {
	for (a=1; a<=R; a++) {
	    tmp = 0.0;
	    for (u=1; u<=M; u++ ) {
	        if (*(WH+offs(N,M,i,u)) > DIFF_LIMIT ) {
		    tmp += *(V+offs(N,M,i,u)) / *(WH+offs(N,M,i,u)) * *(H+offs(R,M,a,u));
#ifdef Isnan
                    if (isnan(tmp)) {
		      fprintf(stderr, "NaN at i%d,a%d,u%d,V%g,WH%g,H%g in updateW()\n", i,a,u, *(V+offs(N,M,i,u)), *(WH+offs(N,M,i,u)), *(H+offs(R,M,a,u)));
		      exit (EXIT_FAILURE);
		    }
#endif
		}
	    }
	    *(W+offs(N,R,i,a)) *= tmp;
	    if ( isnan(*(W+offs(N,R,i,a))) ) {
                fprintf(stderr,"NaN at updateW(i=%d,a=%d)=%g\r",i,a,tmp);
	        exit (EXIT_FAILURE);
	    }
	}
    }
    normalize_by_column(W,N,R);
    return OK;
}

/* update H */
int updateH(void)
{
    int i,u,a;
    double tmp;

    for (a=1; a<=R; a++) {
	for (u=1; u<=M; u++) {
	    tmp = 0.0;
	    for (i=1; i<=N; i++) {
		if ( *(WH+offs(N,M,i,u)) > DIFF_LIMIT ) {
		    tmp += *(W+offs(N,R,i,a)) * *(V+offs(N,M,i,u)) / *(WH+offs(N,M,i,u));
#ifdef Isnan
		    if (isnan(tmp)) {
			fprintf(stderr,"NaN at updateH(i=%d,a=%d)=%g\r",i,a,tmp);
			exit (EXIT_FAILURE);
		    }
#endif
		}
	    }
	    *(H+offs(R,M,a,u)) *= tmp;
#ifdef Isnan
	    if (isnan(*(H+offs(R,M,a,u)))) {
		fprintf(stderr,"NaN at updateH(a=%d,u=%d)=%g\r",a,u,tmp);
		exit (EXIT_FAILURE);
	    }
#endif
	}
    }
    return OK;
}

int nmf_main_loop(double *diff_target)
{
    int iter=0;
    /*
    double diff_target, diff_iter;
    */
    double diff_iter;
    
    while(++iter <= ITER_MAX) {
	matcpy(WH,N,M,oldWH);
	updateW();
	updateH();
	calcWH();
	*diff_target = dmatabs(V, N,M, WH, N, M);
	diff_iter = dmatabs(oldWH, N, M, WH, N, M);

	if (((iter - 1) % EPOCH) == 0) {
	    fprintf(OUT,"### Iteration = %5d, diff_target=%.5g, diff_prev=%g\n", iter, *diff_target, diff_iter); fflush(OUT);
	    if (OUT != stdout) {
		fprintf(stderr,"### Iteration = %5d, diff_target=%.5g, diff_prev=%g\n", iter, *diff_target, diff_iter);
	    }

	    if (verbose) {
		fwritemat(OUT, W,N,R, "### matrix W\n", ""); fflush(OUT);
		fwritemat(OUT, H,R,M, "### matrix H\n", ""); fflush(OUT);
		fwritemat(OUT, WH,N,M, "### matrix WH\n", ""); fflush(OUT);
	    }
	}

	if (*diff_target < ( DIFF_LIMIT * N * M )) {
	    fprintf(OUT,"### Converged at target approximation\n");
	    if (OUT != stdout) {
		fprintf(stderr,"### Converged at target approximation\n");
	    }
	    return iter;
	}
	if (diff_iter < (DIFF_LIMIT * N * M)) {
	    fprintf(OUT,"### Converged at difference between two iterations\n");
	    if (OUT != stdout) {
		fprintf(stderr,"### Converged at difference between two iterations\n");
	    }
	    return iter;
	}
    }
    return (iter-1);
}


int main(int argc, char **argv)
{
    int iterations=0;
    char *stmp=NULL;
    double diff=0.0;

    ParseArgs(argc, argv);
    init_random(seed); seed=get_seed();

    initialize_nmf();

    if ( R > N ) R = N;

    calcWH();

    OUT=fileio_util_openw(output_file_name);
    stmp=malloc(sizeof(char) * (strlen(output_file_name)+1));
    if (stmp==NULL){
	fprintf(stderr, "Could not allocate memory for stmp in main()\n");
	exit (EXIT_FAILURE);
    }
    if ( sprintf(stmp, "%s%c", output_file_name, '\0') <= 0 ) {
	fprintf(stderr, "Could not strcpy in main()\n");
    }

    fprintf(OUT, "### ProgramName: %s --R %d --input %s --iter_max %d --limit %g --epoch %d --seed %d --output %s\n", argv[0], R, data_file_name, ITER_MAX, DIFF_LIMIT, EPOCH, seed, (output_file_name == NULL) ? "-" : output_file_name);
    fprint_signature(OUT); fflush(OUT);

    if (verbose) {
	fwritemat(OUT, W,N,R,  "### initial Weights \n", "");     fflush(OUT);
	fwritemat(OUT, H,R,M,  "### initial Hidden units\n", ""); fflush(OUT);
	fwritemat(OUT, V,N,M,  "### target matrix V\n", "");      fflush(OUT);
	fwritemat(OUT, WH,N,M, "### initial WH\n", "");           fflush(OUT);
    }
    
    iterations = nmf_main_loop(&diff);

    if (iterations == ITER_MAX ) {
	fprintf(OUT, "### did not converge within %d times!!!\n", ITER_MAX);
    } else {
	fprintf(OUT, "### converged at %d times\n", iterations);
    }
    fprintf(OUT, "### total difference between target and estimated matrices=%g, avg=%g\n", diff, diff/((double)N * (double)M));
    
    fwritemat(OUT, W,N,R, "### Weights\n", "");           fflush(OUT);
    fwritemat(OUT, H,R,M, "### Hidden units\n", "");      fflush(OUT);
    fwritemat(OUT, WH,N,M, "### Estimated matrix\n", ""); fflush(OUT);
    fileio_util_close(OUT);
    /*
    if (output_file_name == NULL) {
	return OK;
    }
    */
    if (sprintf(output_file_name,"%s-%05d.H%c", stmp, getpid(), '\0') <= 0) {
	fprintf(stderr, "Can not open file %s\n", output_file_name);
	exit (EXIT_FAILURE);
    }
    OUT=fileio_util_openw(output_file_name);
    fwritemat(OUT, H,R,M, "### Hidden units\n", ""); fflush(OUT);
    fileio_util_close(OUT);
	
    if (sprintf(output_file_name,"%s-%05d.W%c", stmp, getpid(), '\0') <= 0) {
        fprintf(stderr, "Can not open file %s\n", output_file_name);
        exit (EXIT_FAILURE);
    }
    OUT=fileio_util_openw(output_file_name);
    fwritemat(OUT, W,N,R, "### Weights\n", ""); fflush(OUT);
    fileio_util_close(OUT);

    fprintf(stderr, "### total difference between target and estimated matries=%g, avg=%g\n", diff, diff/((double)N * (double)M));
    return OK;
}
